#!/usr/bin/python3
{{/*
Copyright 2020 Mirantis Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/}}

r"""
    The script provides tooling for resque operations (e.g data backup, data restore) procedure
    for mariadb galera cluster (master-master) in k8s/helm environment.
    Currently supported:
        - physical data backup and restore using mariabackup.

    General backup workflow looks like:
        1. Get mariadb pods in ready state and perform simple sanity checks ( e.g check wsrep status
           for each pod).
        2. Select target replica which should be backed up.
        3. Get information needed for backup (target replica ip, persistent volume claim name)
        4. Desynchronize replica from galera cluster.
        5. Launch separate pod (resque runner) with mounted persistent volume from step 3. This pod will
           launch backup commands.
        6. Once resque runner pod is completed, gather logs and remove resque runner pod.
        7. Put mariadb replica back to sync

    General restore workflow looks like:
        1. Get list of mariadb pvcs
        2. Scale mariadb statefulset to 0 replicas
        3. Sequentally for each pvc in list (from step 1) launch a resque runner pod, which
           will restore data to selected pvc from backup volume. After that all pvcs will contain
           identical data.
        4. Scale mariadb statefulset back to MARIADB_REPLICAS replicas

"""

import argparse
import json
import logging
import operator
import os
import sys
import time
import traceback
import jinja2
import pymysql
import pykube
from datetime import datetime
from urllib.parse import urlencode
from waiter import wait
import yaml


def backup_name_format(backup_name):
    names = backup_name.split("/")
    if len(names) > 2:
        raise argparse.ArgumentTypeError(
            f"""Got {backup_name} but backup name should be in <base_backup> or
              <base_backup>/<incremental_backup> format"""
        )
    else:
        for t in names:
            try:
                datetime.strptime(t, "%Y-%m-%d_%H-%M-%S")
            except Exception as e:
                raise argparse.ArgumentTypeError(
                    f"In {backup_name} part {t} should be in %Y-%m-%d_%H-%M-%S format"
                )
    return backup_name


def set_args():
    parser = argparse.ArgumentParser(description="Process resque script arguments")

    subparsers = parser.add_subparsers(
        help="Parse subcommands of resque script", dest="operation"
    )

    for op in ["backup", "restore"]:
        subparser = subparsers.add_parser(
            op, formatter_class=argparse.RawTextHelpFormatter
        )
        if op == "backup":
            subparser.add_argument(
                "--backup-type",
                choices=["incremental", "full"],
                required=True,
                help="""String. Type of backup.
                    incremental - if latest full backup older than MARIADB_FULL_BACKUP_CYCLE make full backup, else
                                make incremental backup
                    full - make full backup""",
            )
            subparser.add_argument(
                "--backup-timeout",
                default=21600,
                type=int,
                help="Integer. Timeout in seconds for backup runner pod to succeed.",
            )
        if op == "restore":
            subparser.add_argument(
                "--backup-name",
                type=backup_name_format,
                required=True,
                help="""String. Name of folder with backup e.g 2020-07-30_10-00-17.
                    Will be restored base backup 2020-07-30_10-00-17. In order to restore to specific
                    incremental backup of base 2020-07-30_10-00-17, specify incremental backup after slash
                    e.g. 2020-07-30_10-00-17/2020-07-30_10-00-36. In this case if 2020-07-30_10-00-36
                    is not found the restore procedure will fail.""",
            )
            subparser.add_argument(
                "--replica-restore-timeout",
                type=int,
                default=3600,
                help="Integer. Timeout in seconds for 1 replica data to be restored to mysql data folder",
            )
    return parser.parse_args()


def login():
    config = pykube.KubeConfig.from_env()
    client = pykube.HTTPClient(config)
    logger.info(f"Created k8s api client from context {config.current_context}")
    return client


def get_env_var(env_var):
    if env_var in os.environ:
        return os.environ[env_var]

    logger.critical(f"environment variable {env_var} not set")
    raise RuntimeError("FATAL")

class Pod(pykube.Pod):

    def logs(
         self,
         container=None,
         pretty=None,
         previous=False,
         since_seconds=None,
         since_time=None,
         timestamps=False,
         tail_lines=None,
         limit_bytes=None,
         stream=False,
     ):
         """
         Produces the same result as calling kubectl logs pod/<pod-name>.
         Check parameters meaning at
         http://kubernetes.io/docs/api-reference/v1/operations/,
         part 'read log of the specified Pod'. The result is plain text.
         """
         log_call = "log"
         params = {}
         if container is not None:
             params["container"] = container
         if pretty is not None:
             params["pretty"] = pretty
         if previous:
             params["previous"] = "true"
         if since_seconds is not None and since_time is None:
             params["sinceSeconds"] = int(since_seconds)
         elif since_time is not None and since_seconds is None:
             params["sinceTime"] = since_time
         if timestamps:
             params["timestamps"] = "true"
         if tail_lines is not None:
             params["tailLines"] = int(tail_lines)
         if limit_bytes is not None:
             params["limitBytes"] = int(limit_bytes)

         query_string = urlencode(params)
         log_call += f"?{query_string}" if query_string else ""
         kwargs = {
             "version": self.version,
             "namespace": self.namespace,
             "operation": log_call,
             "stream": stream,
         }
         r = self.api.get(**self.api_kwargs(**kwargs))
         r.raise_for_status()
         if stream:
             return r

         return r.text

class Settings:
    def __init__(self, operation):
        mariadb_pod_namespace = get_env_var("MARIADB_POD_NAMESPACE")

        # common settings
        self.mariadb_admin_password = get_env_var("MARIADB_PASSWORD")
        self.mariadb_replicas = int(get_env_var("MARIADB_REPLICAS"))
        self.runner_image = get_env_var("MARIADB_RESQUE_RUNNER_IMAGE")
        self.backup_pvc_name = get_env_var("MARIADB_BACKUP_PVC_NAME")
        self.runner_service_account = get_env_var(
            "MARIADB_RESQUE_RUNNER_SERVICE_ACCOUNT"
        )
        self.runner_pod_name_prefix = os.environ.get(
            "MARIADB_RESQUE_RUNNER_POD_NAME_PREFIX", "mariadb"
        )
        self.runner_deletion_timeout = int(
            os.environ.get("MARIADB_RESQUE_RUNNER_DELETION_TIMEOUT", 60)
        )

        self.mariadb_pod_selector = {"application": "mariadb", "component": "server"}
        self.runner_pod_selector = {
            "application": f"mariadb-phy-{operation}-runner",
            "component": f"{operation}-runner",
        }
        self.admin_config_file = "/etc/mysql/admin_user.cnf"

        self.namespace = mariadb_pod_namespace
        self.release_name = f"{mariadb_pod_namespace}-mariadb"

        # backup settings
        if operation == "backup":
            self.full_backup_cycle = int(
                os.environ.get("MARIADB_FULL_BACKUP_CYCLE", 604800)
            )
            self.backups_to_keep = int(os.environ.get("MARIADB_BACKUPS_TO_KEEP", 3))
            self.required_space_ratio = float(
                os.environ.get("MARIADB_BACKUP_REQUIRED_SPACE_RATIO", 1.2)
            )
        # restore settings
        if operation == "restore":
            self.mariadb_sts_name = "mariadb-server"
            self.mariadb_sts_scale_timeout = int(
                os.environ.get("MARIADB_STS_SCALE_TIMEOUT", 800)
            )
            self.runner_restore_node_selector = json.loads(
                get_env_var("MARIADB_RESQUE_RUNNER_RESTORE_NODE_SELECTOR")
            )

    def get_opts_dict(self):
        return self.__dict__


class MysqlBackupException(Exception):
    pass


def mysql_connect(
    host=None,
    port=None,
    unix_socket=None,
    user=None,
    password=None,
    config_file="",
    ssl_cert=None,
    ssl_key=None,
    ssl_ca=None,
    db=None,
    connect_timeout=30,
):
    config = {}

    if ssl_ca is not None or ssl_key is not None or ssl_cert is not None:
        config["ssl"] = {}

    if os.path.exists(config_file):
        config["read_default_file"] = config_file

    if unix_socket is not None:
        config["unix_socket"] = unix_socket
    else:
        if host is not None:
            config["host"] = host
        if port is not None:
            config["port"] = port

    if user is not None:
        config["user"] = user
    if password is not None:
        config["passwd"] = password
    if ssl_cert is not None:
        config["ssl"]["cert"] = ssl_cert
    if ssl_key is not None:
        config["ssl"]["key"] = ssl_key
    if ssl_ca is not None:
        config["ssl"]["ca"] = ssl_ca
    if db is not None:
        config["db"] = db
    if connect_timeout is not None:
        config["connect_timeout"] = connect_timeout

    db_connection = pymysql.connect(**config)

    return db_connection.cursor()


def mysql_get_global_status_vars(host, vars):
    try:
        cursor = mysql_connect(host=host,
            password=SETTINGS.mariadb_admin_password,
            config_file=SETTINGS.admin_config_file)
        cursor.execute(
            "SHOW GLOBAL STATUS WHERE Variable_name IN ({0})".format(
                ",".join(["'" + var + "'" for var in vars])
            )
        )
        val = cursor.fetchall()
        # fetchall returns tuple of tuples, set is needed to compare statuses
        return set(val)
    except Exception as e:
        logger.critical(f"Failed to get mysql global status variables")
        raise e
    finally:
        cursor.close()


def mysql_set_global_variable(host, var, value):
    try:
        cursor = mysql_connect(host=host,
            password=SETTINGS.mariadb_admin_password,
            config_file=SETTINGS.admin_config_file)
        cursor.execute("SET GLOBAL {0} = {1}".format(var, value))
    except Exception as e:
        logger.critical(f"Failed to set mysql variable {var} to value {value}")
        raise e
    finally:
        cursor.close()


def get_pods_query(namespace, selector):
    return Pod.objects(K8S_API).filter(namespace=namespace, selector=selector,)


def get_pods_list(namespace, selector):
    return [pod for pod in get_pods_query(namespace, selector)]


def get_mariadb_ready_pods():
    pods = get_pods_list(SETTINGS.namespace, SETTINGS.mariadb_pod_selector)
    return [pod for pod in filter(operator.attrgetter("ready"), pods)]


def get_pods_logs(namespace, selector):
    try:
        pods = get_pods_list(namespace, selector)
        logger.info(f"Gathering pods logs")
        for pod in pods:
            for container in pod.obj["spec"]["containers"]:
                logger.info(f"Pod {pod.name} {container['name']} LOGS START")
                container_log = pod.logs(
                    container=container["name"], timestamps=True, stream=True,
                )
                with container_log as r:
                    for chunk in r.iter_content(chunk_size=8192):
                        if chunk:
                            ch.stream.buffer.write(chunk)
                logger.info(f"Pod {pod.name} {container['name']} LOGS END")
    except Exception as e:
        logger.critical(f"Got an error while gathering logs")
        logging.exception(e)


def wait_pods_removed(namespace, selector, timeout=60):
    try:
        pods = get_pods_list(namespace, selector)
        for pod in pods:
            logger.info(f"Removing pod {pod.name}")
            pod.delete()
        for _ in wait(10, timeout):
            res = get_pods_list(namespace, selector)
            if not res:
                logger.info(f"Pods {pods} are not present.")
                return
    except Exception as e:
        logger.critical("Unable to remove pods, got an error")
        logging.exception(e)


def get_galera_member_state(host, state_fields):
    try:
        status = mysql_get_global_status_vars(host, state_fields)
        return status
    except Exception as e:
        logger.critical(f"Unable to get galera memeber {host} state")
        raise e


def galera_members_in_state(state, pods):
    statuses = {}
    expected = set(state.items())
    res = True
    for pod in pods:
        host = pod.obj["status"]["podIP"]
        status = get_galera_member_state(host, state.keys())
        diff = status.difference(expected)
        statuses[pod] = {"status": status, "diff": diff}
        if diff:
            res = False
    logger.info(f"Galera members states are {statuses}")
    return res


def pre_backup_sanity_check(pods, synced):
    pods_count = len(pods)
    if pods_count == SETTINGS.mariadb_replicas and galera_members_in_state(
        synced, pods
    ):
        logger.info("Sanity check successful")
    else:
        logger.critical(
            f"Something wrong with Galera cluster. Ready pods count {pods_count}."
        )
        raise RuntimeError("FATAL: Sanity check failed")


def select_target_replica(pods):
    # We need to choose mariadb replica which is not active in load balancer.
    # As nginx ingress controller will be dropped in future, for now consider
    # replica with highest number in name as not active replica.
    def _get_int_pod_suffix(pod_name):
        return int(pod_name.split("-")[-1])

    pod_names = []
    for pod in pods:
        pod_names.append(pod.name)
    target_pod_name = sorted(pod_names, key=_get_int_pod_suffix)[-1]
    for pod in pods:
        if pod.name == target_pod_name:
            return pod


def get_target_replica_info(pod):
    for volume in pod.obj["spec"]["volumes"]:
        if volume["name"] == "mysql-data":
            return (
                pod.name,
                pod.obj["status"]["podIP"],
                pod.obj["spec"]["nodeName"],
                volume["persistentVolumeClaim"]["claimName"],
            )
    raise RuntimeError(
        f"FATAL: Pod doesn't have mysql-data volume, and cannot be target"
    )


def wait_mariadb_desynced(host, pod_name, desynced, timeout):
    for _ in wait(30, timeout):
        logger.info("Waiting galera to become desynced")
        pod = get_pods_query(SETTINGS.namespace, SETTINGS.mariadb_pod_selector).get(
            name=pod_name
        )
        if not pod.ready:
            if galera_members_in_state(desynced, [pod]):
                logger.info("Galera cluster member is desynced.")
                return
    logger.critical("Galera cluster is not in expected state! Exiting.")
    raise RuntimeError("FATAL: failed wait for mariadb desynced")


def wait_mariadb_synced(host, pod_name, synced, timeout):
    for _ in wait(30, timeout):
        logger.info("Waiting galera to become synced")
        pod = get_pods_query(SETTINGS.namespace, SETTINGS.mariadb_pod_selector).get(
            name=pod_name
        )
        if pod.ready:
            if galera_members_in_state(synced, [pod]):
                logger.info("Galera cluster member is synced.")
                return
    logger.critical("Galera cluster is not in expected state! Exiting.")
    raise RuntimeError("FATAL: failed to wait for mariadb synced")


def get_statefulset(name, namespace, selector):
    return (
        pykube.StatefulSet.objects(K8S_API)
        .filter(namespace=namespace, selector=selector,)
        .get_by_name(name)
    )


def get_pvc_query(namespace, selector):
    return pykube.PersistentVolumeClaim.objects(K8S_API).filter(
        namespace=namespace, selector=selector,
    )


def get_pvc_list(namespace, selector):
    return [pvc for pvc in get_pvc_query(namespace, selector)]


def wait_mariadb_sts_scaled(replicas, timeout):
    sts = get_statefulset(
        SETTINGS.mariadb_sts_name, SETTINGS.namespace, SETTINGS.mariadb_pod_selector
    )
    for _ in wait(30, timeout):
        sts.reload()
        sts_status = sts.obj["status"]
        logger.info(
            f"Waiting StatefulSet {SETTINGS.mariadb_sts_name} to be scaled, current sts status is {sts_status}"
        )
        readyReplicas = int(sts_status.get("readyReplicas", 0))
        currentReplicas = int(sts_status.get("currentReplicas", 0))
        sts_replicas = int(sts_status.get("replicas", 0))
        if readyReplicas == currentReplicas == replicas == sts_replicas:
            logger.info(f"StatefulSet {SETTINGS.mariadb_sts_name} scaled")
            return
    logger.critical(
        f"StatefulSet {SETTINGS.mariadb_sts_name} failed to scale within timeout {timeout}"
    )
    raise RuntimeError("FATAL: failed to wait for mariadb StatefulSet scaled")


def get_operation_cmd_args(operation, **kwargs):
    cmd_args_mapping = {
        "backup": [
            "backups_to_keep",
            "full_backup_cycle",
            "backup_type",
            "mariadb_host",
            "required_space_ratio",
        ],
        "restore": ["backup_name", "cleanup_unarchieved_data",],
    }
    cmd_args = []
    for arg in cmd_args_mapping[operation]:
        cmd_args.append(f"--{arg.replace('_', '-')}={kwargs[arg]}")
    return cmd_args


def run_mariadb_operation(operation, timeout, **kwargs):
    try:
        ENV = jinja2.Environment(
            loader=jinja2.FileSystemLoader("/tmp"),
            extensions=["jinja2.ext.do", "jinja2.ext.loopcontrols"],
        )
        tpl = ENV.get_template("resque_runner.yaml.j2")
        logger.debug(f"Using template {tpl.filename}")
        kwargs["cmd_args"] = get_operation_cmd_args(operation, **kwargs)
        text = tpl.render(
            operation=operation,
            timestamp=int(time.time()),
            cmd_args=kwargs["cmd_args"],
            pvc_name=kwargs["pvc_name"],
            namespace=kwargs["namespace"],
            backup_pvc_name=kwargs["backup_pvc_name"],
            runner_pod_name_prefix=kwargs["runner_pod_name_prefix"],
            runner_pod_selector=kwargs["runner_pod_selector"],
            runner_node_selector=kwargs["runner_node_selector"],
            runner_service_account=kwargs["runner_service_account"],
            runner_image=kwargs["runner_image"],
        )
        data = yaml.safe_load(text)
        Pod(K8S_API, data).create()
        runner_status = None
        r_pods = []
        for _ in wait(60, timeout):
            r_pods = get_pods_list(kwargs["namespace"], kwargs["runner_pod_selector"])
            if r_pods:
                runner_status = r_pods[0].obj["status"]
                if runner_status["phase"] == "Failed":
                    logger.critical(
                        f"{operation.upper()} failed! {operation} runner pod status {runner_status}"
                    )
                    raise MysqlBackupException()
                elif runner_status["phase"] == "Succeeded":
                    logger.info(
                        f"{operation.upper()} is finished successfully, {operation} runner pod status {runner_status}"
                    )
                    return
                else:
                    logger.info(
                        f"{operation.upper()} is still running, {operation} runner pod status {runner_status}"
                    )
            else:
                logger.info(f"No any {operation} runner pods found, waiting...")
        logger.critical(
            f"{operation.upper()} is timed out! {operation} runner pod status {runner_status}"
        )
    except Exception as e:
        logger.critical(f"FATAL: {operation} failed")
        logging.exception(e)
        raise e
    finally:
        get_pods_logs(kwargs["namespace"], kwargs["runner_pod_selector"])
        wait_pods_removed(
            kwargs["namespace"],
            kwargs["runner_pod_selector"],
            SETTINGS.runner_deletion_timeout,
        )


def run_mariadb_phy_backup(pod_ip, pvc_name, node_name, timeout, **kwargs):
    _backup_kwargs = {
        "mariadb_host": pod_ip,
        "pvc_name": pvc_name,
        # for backup runner_default_node_selector should be overriden because
        # for example ceph volumes cannot be mounted from different nodes
        "runner_node_selector": {"kubernetes.io/hostname": node_name},
    }
    kwargs.update(_backup_kwargs)

    run_mariadb_operation("backup", timeout, **kwargs)


def run_mariadb_phy_restore(mariadb_pvc_list, timeout, **kwargs):

    pvc_restore_kwargs = {}

    for pvc in mariadb_pvc_list:
        # after restoring last mariadb, remove all unarchieved data
        # from backup pvc
        cleanup_unarchieved_data = False
        if pvc.name == mariadb_pvc_list[-1].name:
            cleanup_unarchieved_data = True

        pvc_restore_kwargs["pvc_name"] = pvc.name
        pvc_restore_kwargs["cleanup_unarchieved_data"] = cleanup_unarchieved_data

        kwargs.update(pvc_restore_kwargs)
        run_mariadb_operation("restore", timeout, **kwargs)


def scale_mariadb_sts(sts, replicas, retries=5):
    # The object can be changed between reload and scale steps,
    # this can cause scale to fail, so adding retries
    res = False
    for i in range(0, retries):
        try:
            sts.reload()
            sts.scale(replicas)
            res = True
        except Exception as e:
            logger.critical(f"{sts.kind} {sts.name} scale failed, retrying")
            logging.exception(e)
            continue
        break
    if not res:
        logger.critical(f"Unable to scale {sts.kind} {sts.name}")
        raise RuntimeError("FATAL")


args = set_args()

operation = args.operation
if not operation:
    parser.error("Mandatory to specify operation")

logger = logging.getLogger(f"mariadb_{operation}")
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)

K8S_API = login()

SETTINGS = Settings(operation)

operation_kwargs = SETTINGS.get_opts_dict()


def main():
    # Concurrent operations are not allowed
    existing_runners = get_pods_list(SETTINGS.namespace, SETTINGS.runner_pod_selector)
    if existing_runners:
        logger.critical(
            f"Found another {operation} already running! Found existing runner pods {existing_runners}"
        )
        raise RuntimeError("FATAL")

    if operation == "backup":
        operation_kwargs["backup_type"] = args.backup_type

        synced_state = {
            "wsrep_ready": "ON",
            "wsrep_cluster_status": "Primary",
            "wsrep_local_state_comment": "Synced",
            "wsrep_cluster_size": str(SETTINGS.mariadb_replicas),
            "wsrep_connected": "ON",
        }
        desynced_state = synced_state.copy()
        desynced_state["wsrep_local_state_comment"] = "Donor/Desynced"

        ready_pods = get_mariadb_ready_pods()
        logger.info(f"Found mariadb ready pods {ready_pods}")

        pre_backup_sanity_check(ready_pods, synced_state)

        target_replica = select_target_replica(ready_pods)
        logger.info(f"Target replica is: {target_replica}")

        (
            mariadb_pod_name,
            mariadb_pod_ip,
            mariadb_pod_node_name,
            mariadb_pod_pvc_name,
        ) = get_target_replica_info(target_replica)

        logger.info(
            f"""Target replica ip is: {mariadb_pod_ip},
                target replica data volume claim is {mariadb_pod_pvc_name}
                target replica node name is {mariadb_pod_node_name}
                """
        )

        try:
            mysql_set_global_variable(mariadb_pod_ip, "wsrep_desync", "ON")
            wait_mariadb_desynced(
                mariadb_pod_ip, mariadb_pod_name, desynced_state, 300
            )

            run_mariadb_phy_backup(
                mariadb_pod_ip,
                mariadb_pod_pvc_name,
                mariadb_pod_node_name,
                args.backup_timeout,
                **operation_kwargs,
            )
        except Exception as e:
            logger.critical("Fatal, got an error")
            raise e
        finally:
            mysql_set_global_variable(mariadb_pod_ip, "wsrep_desync", "OFF")
            wait_mariadb_synced(mariadb_pod_ip, mariadb_pod_name, synced_state, 300)

    elif operation == "restore":
        operation_kwargs["backup_name"] = args.backup_name
        operation_kwargs["runner_node_selector"] = SETTINGS.runner_restore_node_selector

        mariadb_sts = get_statefulset(
            SETTINGS.mariadb_sts_name, SETTINGS.namespace, SETTINGS.mariadb_pod_selector
        )
        mariadb_pvc_list = get_pvc_list(
            SETTINGS.namespace, SETTINGS.mariadb_pod_selector
        )

        scale_mariadb_sts(mariadb_sts, 0)
        wait_mariadb_sts_scaled(0, SETTINGS.mariadb_sts_scale_timeout)

        run_mariadb_phy_restore(
            mariadb_pvc_list, args.replica_restore_timeout, **operation_kwargs,
        )

        scale_mariadb_sts(mariadb_sts, SETTINGS.mariadb_replicas)
        wait_mariadb_sts_scaled(
            SETTINGS.mariadb_replicas, SETTINGS.mariadb_sts_scale_timeout
        )

    logger.info(f"{operation} finished successful!")


if __name__ == "__main__":
    main()
