#!/bin/bash

set -exo pipefail

# Display Help
function help {
cat << EOF
    The script provides tooling for resque operations (e.g data backup, data restore) of mariadb
    galera cluster member.
    Currently supported:
        - physical data backup and restore using mariabackup.
    Example of resulting backup directory structure:

    /var/backup
    |-- base
    |   |-- 2020-07-30_10-00-07
    |   |-- 2020-07-30_10-00-17
    |   |-- 2020-07-30_10-18-43
    |-- incr
    |   |-- 2020-07-30_10-00-17
    |       |-- 2020-07-30_10-00-36
    |       |-- 2020-07-30_10-03-37

    Usage:
      backup [--mariadb-host=<host>]           - ip address or host name of mariadb server (default localhost)
             [--required-space-ratio=<number>] - is a multiplier (floating number e.g 1.2) for predicting space needed
                                                 to create backup and then to do a restore keeping uncompressed backup files
                                                 on the same filesystem as compressed ones. To estimate how big REQUIRED_SPACE_RATIO the next
                                                 formula can be used:
                                                   size of (1 uncompressed full backup + all related incremental uncompressed backups + 1
                                                   full compressed backup) in KB =< (DB_SIZE * REQUIRED_SPACE_RATIO) in KB
                                                 (default 1.2)

             [--full-backup-cycle=<number>]    - number of seconds between full backups, if --backup-type=incremental (default 604800)

             [--backups-to-keep=<number>]      - how much full backups to keep, if there are more than backups-to-keep backups present,
                                                 all oldest full backups will be removed including related incremental backups (default 3)

             [--backup-type=(full|incremental] - whether do full or incremental backup. If incremental is set, full backup will be performed if
                                                 latest full backup is older than full-backup-cycle seconds (default full)

             [--mariadb-client-conf=<path>]    - path on file system to file with mysql client settings (user, password) (default /etc/mysql/mariabackup_user.cnf)

      restore --backup-name=<name>                      - (required) name of directory with full backup to restore e.g 2020-07-29_10-31-52, all
                                                          related incremental backups also have full backup name in the path. In order to restore
                                                          specific incremental backup (and all previous incrementals) specify it after slash e.g:
                                                          2020-07-29_10-31-52/2020-07-30_10-31-52.

              [--cleanup-unarchieved-data=(True|False)] - whether to cleanup unarchieved backup data after restore (default False)

              [--mariadb-client-conf=<path>]            - path on file system to file with mysql client settings (user, password) (default /etc/mysql/mariabackup_user.cnf)

      -h - output this help message
EOF
}

# Prints line number and "message" in error format
# err $LINENO "message"
function err {
    local exitcode=$?
    local xtrace
    xtrace=$(set +o | grep xtrace)
    set +o xtrace
    local msg="[ERROR] ${BASH_SOURCE[2]}:$1 $2"
    echo "$msg" 1>&2;
    $xtrace
    return $exitcode
}

# Prints backtrace info
# filename:lineno:function
# backtrace level
function backtrace {
    local level=$1
    local deep
    deep=$((${#BASH_SOURCE[@]} - 1))
    echo "[Call Trace]"
    while [[ ${level} -le ${deep} ]]; do
        echo "${BASH_SOURCE[$deep]}:${BASH_LINENO[$deep-1]}:${FUNCNAME[$deep-1]}"
        deep=$((deep - 1))
    done
}

# Prints line number and "message" then exits
# die $LINENO "message"
function die {
    local exitcode=$?
    set +o xtrace
    local line=$1; shift
    if [[ "${exitcode}" == 0 ]]; then
        exitcode=1
    fi
    backtrace 2
    err "${line}" "$*"
    # Give buffers a second to flush
    sleep 1
    exit $exitcode
}

# checks that arg 1 match regex 2
function check_optional_arg {
    if [[ ! ${1} =~ ${2} ]]; then
        help
        die $LINENO "Some parameter is set to incorrect value."
    fi
}


OPERATION=$1

case $OPERATION in
    restore)
    echo "Running mariadb physical restore"
    shift
    for i in "$@"; do
        case $i in
            --cleanup-unarchieved-data=*)
            CLEANUP_UNARCHIEVED_DATA="${i#*=}"
            check_optional_arg "${CLEANUP_UNARCHIEVED_DATA}" "(^True$|^False$)"
            shift
            ;;
            --backup-name=*)
            MARIADB_BACKUP_NAME="${i#*=}"
            if [[ ! "${MARIADB_BACKUP_NAME}" ]]; then
                die $LINENO "Target backup name is undefined"
            fi
            shift
            ;;
            --mariadb-client-conf=*)
            MARIADB_CLIENT_CONF="${i#*=}"
            shift
            ;;
            *)
            help
            die $LINENO "Unsupported restore option"
            ;;
        esac
    done
    ;;
    backup)
    echo "Running mariadb physical backup"
    shift
    for i in "$@"; do
        case $i in
            --mariadb-host=*)
            MARIADB_HOST="${i#*=}"
            shift
            ;;
            --required-space-ratio=*)
            REQUIRED_SPACE_RATIO="${i#*=}"
            shift
            ;;
            --full-backup-cycle=*)
            MARIADB_FULL_BACKUP_CYCLE="${i#*=}"
            shift
            ;;
            --backups-to-keep=*)
            MARIADB_BACKUPS_TO_KEEP="${i#*=}"
            shift
            ;;
            --backup-type=*)
            MARIADB_BACKUP_TYPE="${i#*=}"
            check_optional_arg "${MARIADB_BACKUP_TYPE}" "(^full$|^incremental$)"
            shift
            ;;
            --mariadb-client-conf=*)
            MARIADB_CLIENT_CONF="${i#*=}"
            shift
            ;;
            *)
            help
            die $LINENO "Unsupported backup option"
            ;;
        esac
    done
    ;;
    -h)
    help
    exit 0
    ;;
    *)
    help
    die $LINENO "Unsupported operation"
    ;;
esac

function lock {
    if [[ -d "${LOCKDIR}" ]]; then
        die $LINENO "Lock directory ${LOCKDIR} already exists"
    else
        mkdir "${LOCKDIR}" || die $LINENO "Could not create lock directory ${LOCKDIR}"
        echo "Lock directory created. Check completed."
    fi
}

function unlock {
    if [[ -d "${LOCKDIR}" ]]; then
        if rmdir "${LOCKDIR}"; then
            echo "Lock directory ${LOCKDIR} removed"
        else
            die $LINENO "Unable to remove lock directory ${LOCKDIR}"
        fi
    else
       echo "Lock directory ${LOCKDIR} does not exist, nothing to unlock"
    fi
}

function post_operation_info(){
    SPENT=$(( $(date +%s) - START ))
    echo "Took ${SPENT} seconds. Completed: $(date)"
    echo "Backup directories tree:"
    tree -d "${BACKDIR}"
}

function set_global_variables(){
    BACKCMD=mariabackup
    BACKDIR=/var/backup
    LOCKDIR=/var/backup/mariabackup.lock
    MARIADB_CLIENT_CONF=${MARIADB_CLIENT_CONF:-/etc/mysql/mariabackup_user.cnf}
    USEROPTIONS="--defaults-file=${MARIADB_CLIENT_CONF}"
    BASEBACKDIR="${BACKDIR}/base"
    INCRBACKDIR="${BACKDIR}/incr"
    START=$(date +%s)

    if [[ $OPERATION == 'backup' ]]; then
        MARIADB_FULL_BACKUP_CYCLE=${MARIADB_FULL_BACKUP_CYCLE:-604800} # Create a new full backup every X seconds
        MARIADB_BACKUPS_TO_KEEP=${MARIADB_BACKUPS_TO_KEEP:-3} # Number of additional backups cycles a backup should kept for.
        MARIADB_BACKUP_TYPE=${MARIADB_BACKUP_TYPE:-full}
        if [[ "${MARIADB_BACKUP_TYPE}" != "full" &&  -d "${BASEBACKDIR}" ]]; then   # Check if MariaDB upgraded since last backup
            local tools_version=$(get_mariabackup_version)
            local backup_version=$(get_backup_version)
            if [[ "${tools_version}" != "${backup_version}" ]]; then
                echo "Version of MariaDB changed. Process FULL backup instead ${MARIADB_BACKUP_TYPE}."
                MARIADB_BACKUP_TYPE="full"
            fi
        fi
        MARIADB_HOST=${MARIADB_HOST:-localhost}
        USEROPTIONS="${USEROPTIONS} --host=${MARIADB_HOST}"
        REQUIRED_SPACE_RATIO=${REQUIRED_SPACE_RATIO:-'1.2'}
    elif [[ $OPERATION == 'restore' ]]; then
        if [[ "${MARIADB_BACKUP_NAME}" =~ ^.*/.*$ ]]; then
            TARGET_BASE_BACKUP_NAME="$(echo ${MARIADB_BACKUP_NAME} | cut -d/ -f1)"
            TARGET_INCR_BACKUP_NAME="$(echo ${MARIADB_BACKUP_NAME} | cut -d/ -f2)"
        else
            TARGET_BASE_BACKUP_NAME="${MARIADB_BACKUP_NAME}"
        fi
        TARGET_BASE_BACKUP_DIR="${BASEBACKDIR}/${TARGET_BASE_BACKUP_NAME}"
        if [[ ! -d "${TARGET_BASE_BACKUP_DIR}" ]]; then
            die $LINENO "Target backup directory ${TARGET_BASE_BACKUP_DIR} does not exist"
        fi
        local incr_parent_path="${INCRBACKDIR}/${TARGET_BASE_BACKUP_NAME}"

        if [[ "${TARGET_INCR_BACKUP_NAME}" ]]; then
            local incr_dirs
            local incr_parent_path="${INCRBACKDIR}/${TARGET_BASE_BACKUP_NAME}"
            [[ -d "${incr_parent_path}" ]] || die $LINENO "Incremental backups parent directory ${incr_parent_path} not found"
            incr_dirs=$(find ${incr_parent_path} -name backup.stream.gz | xargs dirname | sort -n)
            local target_incr_backup_path="${incr_parent_path}/${TARGET_INCR_BACKUP_NAME}"
            local num

            echo "Restore to specific incremental backup was requested.
                  Searching incremental backup ${target_incr_backup_path} among ${incr_dirs}
                  and all previous incremental backups if any."
            # All backups till target backup and including it
            num=$(echo "${incr_dirs}" | grep -n -E ^"${target_incr_backup_path}"$ | cut -d: -f1)
            [[ "${num}" ]] || die $LINENO "Incremental backup ${target_incr_backup_path} not found!"
            TARGET_INCR_BACKUP_DIRS=$(echo "${incr_dirs}" | head "-${num}")
        fi
        CLEANUP_UNARCHIEVED_DATA=${CLEANUP_UNARCHIEVED_DATA:-'False'}
    fi
}

function pre_backup_sanity_check() {
    local available_space
    local db_size
    local required_space
    mysql $USEROPTIONS -s -e 'exit' || die $LINENO "FATAL ERROR: Could not connect to mysql with provided credentials"
    # check available space on device in KB
    available_space=$(df --output=avail -k ${BACKDIR} | sed 's/[^0-9]*//g' | tr -d '[:space:]')
    # Calculate current db size
    db_size=$(du -s --exclude=galera.cache --exclude=ib_logfile* /var/lib/mysql | sed 's/[^0-9]*//g' | tr -d '[:space:]')
    required_space=$(echo | awk "{ printf \"%.f\", $db_size*$REQUIRED_SPACE_RATIO }")
    if [[ $available_space -lt $required_space ]]; then
        die $LINENO "FATAL ERROR: Not enough space for backup,
                     AVAILABLE_SPACE is ${available_space},
                     DB_SIZE is ${db_size},
                     REQUIRED_SPACE is ${required_space}"
    fi
}

function set_global_backup_cmd_options_and_target_dir() {
    local backup_type
    local latest_base_backup_dir
    local latest_base_backup_date
    local latest_base_backup_age
    local do_new_base_backup
    backup_type=$1
    if [[ "${backup_type}" == "incremental" ]]; then
        latest_base_backup_dir=$(find $BASEBACKDIR -mindepth 1 -maxdepth 1 -type d -printf "%P\n" | sort -nr | head -1)
        if [[ "${latest_base_backup_dir}" ]]; then
            local latest_base_backup_path
            latest_base_backup_path="${BASEBACKDIR}/${latest_base_backup_dir}"
            if [[ ! -f "${latest_base_backup_path}/backup.successful" ]]; then
                die $LINENO "Successful base backup not found at ${latest_base_backup_path}!!!! Cannot proceed"
            fi
            # Getting latest base backup age unix epoch format
            # all backup directories names are in format YYYY-MM-DD_hh-mm-ss,
            # as ubuntu doesn't provide magic option for date so we need to convert
            # this date to supported format e.g. YYYY-MM-DDThh:mm:ss
            latest_base_backup_date=$(echo "${latest_base_backup_dir}" | awk '{ split($0,a,"_"); gsub("-", ":", a[2]); print a[1]"T"a[2]}')
            latest_base_backup_age=$(date -d "${latest_base_backup_date}" "+%s")
            # do incremental backup if type is incremental and there is at list one full backup present and if current
            # date is lower than latest full backup age + MARIADB_FULL_BACKUP_CYCLE, else do full backup
            if [[ $(( latest_base_backup_age + MARIADB_FULL_BACKUP_CYCLE + 5 )) -ge ${START} ]]; then
                local incr_backups_path
                local incremental_basedir
                local latest_incr_backup_dir
                incr_backups_path="${INCRBACKDIR}/${latest_base_backup_dir}"
                incremental_basedir="${latest_base_backup_path}"
                if [[ -d "${incr_backups_path}" ]]; then
                    latest_incr_backup_dir=$(find "${incr_backups_path}" -mindepth 1  -maxdepth 1 -type d -printf "%P\n" | sort -nr | head -1)
                    if [[ "${latest_incr_backup_dir}" ]]; then
                        local latest_incr_backup_path
                        latest_incr_backup_path="${incr_backups_path}/${latest_incr_backup_dir}"
                        if [[ ! -f "${latest_incr_backup_path}/backup.successful" ]]; then
                            die $LINENO "Successful incremental backup at ${latest_incr_backup_path}!!! Cannot proceed"
                        fi
                        # This is a 2+ incremental backup
                        incremental_basedir="${latest_incr_backup_path}"
                        echo "New incremental backup for full backup ${latest_base_backup_dir}, base incremental backup is ${incremental_basedir}"
                    fi
                else
                    echo "The first incremental backup for base backup ${latest_base_backup_dir}"
                fi
                TARGET_DIR="${incr_backups_path}/$(date +%F_%H-%M-%S)"
                # Set incremental Backup options
                BACKUP_OPTIONS="${USEROPTIONS} --backup --extra-lsndir=${TARGET_DIR} --incremental-basedir=${incremental_basedir} --stream=xbstream"
                return
            fi
        fi
        do_new_base_backup="True"
    fi
    if [[ "${backup_type}" == "full" || "${do_new_base_backup}" == "True" ]]; then
        TARGET_DIR="${BASEBACKDIR}/$(date +%F_%H-%M-%S)"
        echo "New full backup to directory ${TARGET_DIR}"
        # Set full backup options
        BACKUP_OPTIONS="${USEROPTIONS} --backup --extra-lsndir=${TARGET_DIR} --stream=xbstream"
    fi
}

function run_backup() {
    mkdir -p "${TARGET_DIR}"

    $BACKCMD $BACKUP_OPTIONS | gzip > "${TARGET_DIR}/backup.stream.gz"
    # backup original grastate.dat to make sure that restore will be done correctly,
    # xtrabackup_galera_info doesn't contain all fields from grastate.dat
    cp /var/lib/mysql/grastate.dat "${TARGET_DIR}/"
    touch "${TARGET_DIR}/backup.successful"
}

function purge_old_backups() {
    local full_backups
    local full_backups_count
    local remove_count="0"
    full_backups="$(find ${BASEBACKDIR} -mindepth 1 -maxdepth 1 -type d -printf '%P\n' | sort -r)"
    full_backups_count=$(echo "${full_backups}" | wc -l)
    remove_count=$(( $full_backups_count - ${MARIADB_BACKUPS_TO_KEEP} ))
    if [[ "$remove_count" -gt "0" ]]; then
       local backups_to_delete
       backups_to_delete=$(echo "${full_backups}" | tail -${remove_count})
       echo "Backups ${backups_to_delete} will be removed
             Keeping ${MARIADB_BACKUPS_TO_KEEP} newest backups
            "
       for DEL in $backups_to_delete; do
           echo "deleting ${DEL}"
           rm -rf "${BASEBACKDIR:?}/${DEL}"
           rm -rf "${INCRBACKDIR:?}/${DEL}"
       done
    else
        echo "Nothing to cleanup"
    fi
}

function decompress() {
    local dst="${1}/unarchieved"
    mkdir "${dst}"
    zcat "${1}/backup.stream.gz" | mbstream -x -C "${dst}"
}

function unarchieve_backups() {
    local threads
    local dirs
    threads="1"
    dirs="${TARGET_BASE_BACKUP_DIR}"
    # only parallelism up to 8 is allowed for decompressing, to avoid risk of extra high load during
    # decompress
    export -f decompress
    if [[ "${TARGET_INCR_BACKUP_DIRS}" ]]; then
        threads=$(( threads + $(echo "${TARGET_INCR_BACKUP_DIRS}" | wc -l) ))
        if [[ $threads -gt 8 ]]; then
            threads=8
        fi
        dirs="${dirs}\n${TARGET_INCR_BACKUP_DIRS}"
    fi
    echo -e "${dirs}" | xargs -n 1 -P ${threads} -I {} bash -c 'set -ex; decompress "$@"' _ {}
}

function prepare_backup_for_restore(){
    # prepare backups, the first directory in BACKUP_DIRS is the directory with full backup, other lines may contain
    # directories with incremental backups
    ${BACKCMD} --prepare --target-dir "${TARGET_BASE_BACKUP_DIR}/unarchieved/"
    for dir in ${TARGET_INCR_BACKUP_DIRS}; do
        echo "Preparing incremental backup ${dir}"
        ${BACKCMD} --prepare --target-dir "${TARGET_BASE_BACKUP_DIR}/unarchieved/" --incremental-dir="${dir}/unarchieved"
    done
    touch "${TARGET_BASE_BACKUP_DIR}/.prepared"
}

function run_restore() {
    # TODO: implement rollback of restore
    # mariabackup requires datadir to be clean
    local latest_grastate_file
    if [[ ! "${TARGET_INCR_BACKUP_DIRS}" ]]; then
        echo "NO incremental backups found for base backup ${TARGET_BASE_BACKUP_DIR}, only base backup will be restored"
        latest_grastate_file=${TARGET_BASE_BACKUP_DIR}/grastate.dat
    else
        latest_grastate_file=$(echo "${TARGET_INCR_BACKUP_DIRS}" | tail -n 1)/grastate.dat
    fi
    rm -rf /var/lib/mysql/*
    ${BACKCMD} ${USEROPTIONS} --copy-back --target-dir "${TARGET_BASE_BACKUP_DIR}/unarchieved/" --datadir /var/lib/mysql
    cp "${latest_grastate_file}" /var/lib/mysql
}

function cleanup_unarchieved_data(){
    rm -f "${TARGET_BASE_BACKUP_DIR}/.prepared"

    for dir in ${TARGET_INCR_BACKUP_DIRS}; do
        rm -rf "${dir}/unarchieved"
    done
    rm -rf "${TARGET_BASE_BACKUP_DIR}/unarchieved"
}

function get_mariabackup_version() {
    echo "$(${BACKCMD} --version 2>&1 | sed -nE 's/.*(10\.[0-9]{1,2}\.[0-9]{1,2}-MariaDB).*/\1/p')"
}

function get_backup_version() {
    local backup=$1
    # if function called without params, we looking for latest full backup and return its version
    if [[ ! ${backup} ]]; then
        backup=$(find ${BASEBACKDIR} -mindepth 1 -name xtrabackup_info -printf "%h\n" | sort | tail -1)
    fi
    if [[ ! ${backup} ]]; then
        echo "No-backups"
    else
        echo "$(cat ${backup}/xtrabackup_info | grep tool_version | cut -d' ' -f3)"
    fi
}

function pre_restore_sanity_check() {
    local tools_version=$(get_mariabackup_version)
    local backup_version=$(get_backup_version "${TARGET_BASE_BACKUP_DIR}")
    if [[ "${tools_version}" != "${backup_version}" ]]; then
        die $LINENO "Backup need ${backup_version} for restore but we have ${tools_version}"
    fi
}

# Exit on any errors so that errors don't compound and make cleanup
trap err_trap EXIT
function err_trap {
    local r=$?
    if [[ $r -ne 0 ]]; then
        set +o xtrace
        post_operation_info
        if [[ $OPERATION == 'backup' ]]; then
            unlock
        elif [[ $OPERATION == 'restore' ]]; then
            cleanup_unarchieved_data
            unlock
        fi
        echo "${0##*/} FAILED"
    fi
    exit $r
}

echo "started: $(date)"
echo

set_global_variables

if [[ $OPERATION == 'backup' ]]; then

    pre_backup_sanity_check
    lock
    mkdir -p $BASEBACKDIR

    set_global_backup_cmd_options_and_target_dir $MARIADB_BACKUP_TYPE

    run_backup

    purge_old_backups

elif [[ $OPERATION == 'restore' ]]; then

    pre_restore_sanity_check
    lock

    if [[ ! -f "${TARGET_BASE_BACKUP_DIR}/.prepared" ]]; then
        unarchieve_backups
        prepare_backup_for_restore
    fi
    run_restore
    # cleanup unarchieved data, only if flag is set, it is needed to be able to
    # restore multiple replicas from one folder
    if [[ "${CLEANUP_UNARCHIEVED_DATA}" == "True" ]]; then
        cleanup_unarchieved_data
    fi
fi

post_operation_info

unlock
